// **********************************************************************
// PUCRS/Escola Politcnica
// COMPUTAO GRFICA
//
// Programa para geracao de terrenos em OpenGL
//
// Marcio Sarroglia Pinho
// pinho@pucrs.br
// **********************************************************************

#include <iostream>
#include <cmath>
#include <ctime>

using namespace std;

#ifdef WIN32
#include <windows.h>
#include "gl\glut.h"
    static DWORD last_idle_time;
#else
    #include <sys/time.h>
    static struct timeval last_idle_time;
#endif

#ifdef __APPLE__
#include <GLUT/glut.h>
#endif

#include "ImageClass.h"

ImageClass ImagemDoTerreno;

class TPoint  // Struct para armazenar um ponto
{
public:
    
    float X,Y,Z;
    TPoint (float x=0, float y=0, float z=0)
    {
        X = x; Y = y; Z = z;
    }
    void Set(float x, float y, float z)
    {
        X = x;
        Y = y;
        Z = z;
    }
    void Imprime()
    {
        cout << "X: " << X << " Y: " << Y << " Z: " << Z;
    }
} ;

// **********************************************************************
class TTriangle // Struct para armazenar um tringulo
{
public:
    TPoint P1, P2, P3;
    TPoint N1, N2, N3; // Normais dos vertices do triangulo
    void imprime()
    {
        cout << "P1 ";  P1.Imprime(); cout << endl;
        cout << "P2 ";  P2.Imprime(); cout << endl;
        cout << "P3 ";  P3.Imprime(); cout << endl;
    }
} ;

// Rotina que faz um produto vetorial
void ProdVetorial (TPoint v1, TPoint v2, TPoint &vresult)
{
    vresult.X = v1.Y * v2.Z - (v1.Z * v2.Y);
    vresult.Y = v1.Z * v2.X - (v1.X * v2.Z);
    vresult.Z = v1.X * v2.Y - (v1.Y * v2.X);
}

// Esta rotina transforma um vetor em unitario
void VetUnitario(TPoint &vet)
{
    float modulo;
    
    modulo = sqrt (vet.X * vet.X + vet.Y * vet.Y + vet.Z * vet.Z);
    
    if (modulo == 0.0) return;
    
    vet.X /= modulo;
    vet.Y /= modulo;
    vet.Z /= modulo;
}
// **********************************************************************
// void DesenhaTriangulo()
// **********************************************************************
void DesenhaTriangulo(TTriangle T)
{
    glBegin ( GL_TRIANGLES );
        glNormal3f(T.N1.X,T.N1.Y,T.N1.Z); glVertex3f(T.P1.X, T.P1.Y, T.P1.Z);
        glNormal3f(T.N2.X,T.N2.Y,T.N2.Z); glVertex3f(T.P2.X, T.P2.Y, T.P2.Z);
        glNormal3f(T.N3.X,T.N3.Y,T.N3.Z); glVertex3f(T.P3.X, T.P3.Y, T.P3.Z);
    glEnd();
}

#define MAX 300
// **********************************************************************
class Terreno{
    int nVerticeX, nVerticeZ;
    float Largura, Profundidade, AlturaMaxima;
    
    TPoint PontosDoTerreno[MAX][MAX];
    TTriangle FacesDoTerreno[MAX+1][(MAX-1)*2];

public:
    Terreno(int nVerticeX, int nVerticeZ, float Largura, float Profundidade, float AlturaMaxima)
    {
        if ((nVerticeX >= MAX) || (nVerticeZ >= MAX))
        {
            cout << "Erro: Dimensao maxima excedida. Verifique a constante MAX." << endl;
            exit(1);
        }
        this->nVerticeX = nVerticeX;
        this->nVerticeZ = nVerticeZ;
        this->Largura = Largura;
        this->Profundidade = Profundidade;
        this->AlturaMaxima = AlturaMaxima;
        
        GeraVerticesDoTerreno(nVerticeX,nVerticeZ);
        GeraTriangulosDoTerreno(nVerticeX,nVerticeZ);
    }
    int getnVerticeX() { return nVerticeX; }
    int  getnVerticeZ() { return nVerticeZ; }
    float getLargura() { return Largura; }
    float getProfundidade() { return Profundidade; }
    float getAlturaMaxima() { return AlturaMaxima; }
    
    // **********************************************************************
    TTriangle getFaceDoTerreno(int x, int z)
    {
        return FacesDoTerreno[x][z];
    }

    // **********************************************************************
    void GeraVerticesDoTerreno(int nVerticeX, int nVerticeZ)
    {
        TPoint meio(nVerticeX/2.0, 0, nVerticeZ/2.0);
        
        float MaxDist = sqrt(meio.X*meio.X + meio.Z*meio.Z);
        float deltaX = Largura/(nVerticeX-1);
        float deltaZ = Profundidade/(nVerticeZ-1);
        for (int x=0;x<nVerticeX;x++)
        {
            for (int z=0;z<nVerticeZ;z++)
            {
                float dx, dz;
                // calcula o centro do terreno
                dx = meio.X-x;
                dz = meio.Z-z;
                // calcula a distancia do vrtice at o centro do terreno
                float distancia = sqrt(dx*dx + dz*dz);
                // Gera uma altura entre 0 e 1. Quanto mais distante do centro, maior o valor
                float y = (MaxDist - distancia)/MaxDist;
                // Calcula a altura final como uma funo da altura mxima
                y *= AlturaMaxima;
                // Calcula o X e o Z em funo da distncia entre cada vrtice no plano XZ
                PontosDoTerreno[x][z] = TPoint(x*deltaX, y, z*deltaZ);
            }
        }
        //AlteraVerticesDoTerreno();
    }
    void AlteraVerticesDoTerreno()
    {
        //cout << "AlteraVerticesDoTerreno" << endl;
        for (int x=0;x<nVerticeX;x++)
        {
            for (int z=0;z<nVerticeZ;z++)
            {
                if ((x<5) || (z < 5))
                    PontosDoTerreno[x][z].Y = 0;
                if (x>(nVerticeX-5) || z >(nVerticeZ-5))
                    PontosDoTerreno[x][z].Y = 0;
            }
        }
    }
    void AlteraVerticesDoTerreno(int xCentro, int zCentro, int altura, int raio)
    {
        int xi = xCentro-raio;
        int xf = xCentro+raio;
        int zi = zCentro-raio;
        int zf = zCentro+raio;
        cout << "xi: " << xi;
        cout << " xf: " << xf;
        cout << " zi: " << zi;
        cout << " zf: " << zf << endl;
        
        if ( (xf >= nVerticeX) || (zf >= nVerticeZ) || (xi < 0 ) || (zi < 0 ))
        {
            cout << "Erro: Area ultrapassa os limites da malha." << endl;
            exit(1);
        }
    }
    
    // **********************************************************************
    void GeraTriangulosDoTerreno(int nVerticeX, int nVerticeZ)
    {
        //cout << "GeraTriangulosDoTerreno" << endl;
        TPoint A,B,C,D;
        for (int x=0;x<nVerticeX-1;x++)
        {
            for (int z=0;z<nVerticeZ-1;z++)
            {
                // ***************************
                // Gera os quatro vertices que compoem a clula (X,Z)
                A = PontosDoTerreno[x][z];
                B = PontosDoTerreno[x][z+1];
                C = PontosDoTerreno[x+1][z+1];
                D = PontosDoTerreno[x+1][z];
                
                // ***************************
                // Monta o primeiro tringulo
                FacesDoTerreno[x][z*2].P1 = A;
                FacesDoTerreno[x][z*2].P2 = B;
                FacesDoTerreno[x][z*2].P3 = C;
                
                // Calcula a normal do primeiro triangulo
                TPoint VAB, VBC, V1;
                VAB.Set(B.X-A.X, B.Y-A.Y, B.Z-A.Z );
                VBC.Set(C.X-B.X, C.Y-B.Y, C.Z-B.Z );
                ProdVetorial(VAB, VBC, V1);
                VetUnitario(V1);
                
                // Armazena as normais de cada vertice. As 3 normais nos vrtices serao iguais
                FacesDoTerreno[x][z*2].N1 = V1;
                FacesDoTerreno[x][z*2].N2 = V1;
                FacesDoTerreno[x][z*2].N3 = V1;

                // ***************************
                // Monta o segundo tringulo
                FacesDoTerreno[x][z*2+1].P1 = A;
                FacesDoTerreno[x][z*2+1].P2 = C;
                FacesDoTerreno[x][z*2+1].P3 = D;
                
                // Calcula a normal do segundo triangulo
                TPoint  VAC, VCD, V2;
                VAC.Set(C.X-A.X, C.Y-A.Y, C.Z-A.Z );
                VCD.Set(D.X-C.X, D.Y-C.Y, D.Z-C.Z);
                ProdVetorial (VAC, VCD, V2);
                VetUnitario(V2);
                // Armazena as normais de cada vertice. As 3 normais nos vrtices serao iguais
                FacesDoTerreno[x][z*2+1].N1 = V2;
                FacesDoTerreno[x][z*2+1].N2 = V2;
                FacesDoTerreno[x][z*2+1].N3 = V2;
            }
            //cout << endl;
        }
        //cout << "----" << endl;
    }
    // **********************************************************************
    void DesenhaTerreno()
    {
        int nTrianglesInZ = (getnVerticeZ()-1)*2;
        for (int x=0; x< getnVerticeX();x++)
        {
            for (int z=0;z<nTrianglesInZ;z++)
            {
                glPushMatrix();
                DesenhaTriangulo(getFaceDoTerreno(x,z));
                glPopMatrix();
            }
        }
    }
}; // Terreno

GLfloat AspectRatio, AngY=0;
// Referencia para um terreno
Terreno *T1;

bool WireFrame=false;

// **********************************************************************
//
// **********************************************************************
void DesenhaPonto(TPoint P)
{
    glBegin(GL_POINTS);
        glVertex3f(P.X, P.Y, P.Z);
    glEnd();
}
// **********************************************************************
//  void DefineLuz(void)
// **********************************************************************
void DefineLuz(void)
{
  GLfloat LuzAmbiente[]   = {0.4, 0.4, 0.4f } ;
  GLfloat LuzDifusa[]   = {0.7, 0.7, 0.7};
  GLfloat LuzEspecular[] = {0.9f, 0.9f, 0.9 };
  GLfloat PosicaoLuz0[]  = {0.0f, 10.0f, 0.0f };  // Posio da Luz
  GLfloat Especularidade[] = {0.3f, 0.3f, 0.3f};

   // ****************  Fonte de Luz 0

 glEnable ( GL_COLOR_MATERIAL );

   // Habilita o uso de iluminao
  glEnable(GL_LIGHTING);

  // Ativa o uso da luz ambiente
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, LuzAmbiente);
  // Define os parametros da luz nmero Zero
  glLightfv(GL_LIGHT0, GL_AMBIENT, LuzAmbiente);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, LuzDifusa  );
  glLightfv(GL_LIGHT0, GL_SPECULAR, LuzEspecular  );
  glLightfv(GL_LIGHT0, GL_POSITION, PosicaoLuz0 );
  glEnable(GL_LIGHT0);

  // Ativa o "Color Tracking"
  glEnable(GL_COLOR_MATERIAL);

  // Define a reflectancia do material
  glMaterialfv(GL_FRONT,GL_SPECULAR, Especularidade);

  // Define a concentraoo do brilho.
  // Quanto maior o valor do Segundo parametro, mais
  // concentrado ser o brilho. (Valores vlidos: de 0 a 128)
  glMateriali(GL_FRONT,GL_SHININESS,50);

}

// **********************************************************************
//  void init(void)
//		Inicializa os parametros globais de OpenGL
// **********************************************************************
void init(void)
{
	glClearColor(0.5f, 0.5f, 0.5f, 1.0f); // Fundo de tela preto

	glShadeModel(GL_SMOOTH);
	glColorMaterial ( GL_FRONT, GL_AMBIENT_AND_DIFFUSE );
	glEnable(GL_DEPTH_TEST);
	glEnable ( GL_CULL_FACE );
    //glEnable(GL_NORMALIZE);

    // Carrega um terreno de teste
    //ImagemDoTerreno.Load("Mapa-B-100x100.jpg");
    // *****************
    // Inicializa os parmetros do terreno
    int DimX = 50; // numero de vrtices ao longo do eixo X
    int DimZ = 50; // numero de vrtices ao longo do eixo Z
 
    float Larg = 10;
    float Prof = 10;
    float AltMax = 3;
    //Terreno(int nVerticeX, int nVerticeZ, float Largura, float Profundidade, float AlturaMaxima)
    T1 = new Terreno (DimX, DimZ, Larg, Prof, AltMax);


    // Obtem o tempo inicial
#ifdef WIN32
    last_idle_time = GetTickCount();
#else
    gettimeofday (&last_idle_time, NULL);
#endif

}
// **********************************************************************
//  void PosicUser()
// **********************************************************************
void PosicUser()
{
	// Define os parmetros da projeo Perspectiva
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50,AspectRatio,0.01,200);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	gluLookAt(0, 10, T1->getProfundidade()*0.8,   // Posio do Observador
              0,0,0,       // Posio do Alvo
			  0.0f,1.0f,0.0f);

}
// **********************************************************************
//  void reshape( int w, int h )
//		trata o redimensionamento da janela OpenGL
// **********************************************************************
void reshape( int w, int h )
{

	// Evita diviso por zero, no caso de uam janela com largura 0.
	if(h == 0) h = 1;
    // Ajusta a relao entre largura e altura para evitar distoro na imagem.
    // Veja funo "PosicUser".
	AspectRatio = 1.0f * w / h;
	// Inicializa a matriz de projecao
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// Seta a viewport para ocupar toda a janela
    glViewport(0, 0, w, h);
	PosicUser();
}

// **********************************************************************
//  void display( void )
// **********************************************************************
void display( void )
{

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	DefineLuz();
	PosicUser();
	
	glMatrixMode(GL_MODELVIEW);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glPushMatrix();
        // Posiciona o terreno de forma que o meio deste fique na posio (0,0,0)
        glTranslatef ( -T1->getLargura()/2, 0.0f, -T1->getProfundidade()/2 );
        glColor3f(0.0f,0.5f,0.0f); // Verde
        T1->DesenhaTerreno();
    
        // Se quiser desenhar as linhas que compoem os triangulos, pressione espaco
        // para mudar o valor da variavel "WireFrame"
        if (WireFrame)
        {
            glColor3f(0.0f,0.0f,0.0f);
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            T1->DesenhaTerreno();
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        }
    glPopMatrix();
	glutSwapBuffers();
}

// **********************************************************************
//  void animate ( unsigned char key, int x, int y )
// **********************************************************************
void animate()
{
    static float dt;
    static float AccumTime=0;
    
#ifdef _WIN32
    DWORD time_now;
    time_now = GetTickCount();
    dt = (float) (time_now - last_idle_time) / 1000.0;
#else
    // Figure out time elapsed since last call to idle function
    struct timeval time_now;
    gettimeofday(&time_now, NULL);
    dt = (float)(time_now.tv_sec  - last_idle_time.tv_sec) +
    1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
#endif
    AccumTime +=dt;
    if (AccumTime >=3) // imprime o FPS a cada 3 segundos
    {
        cout << 1.0/dt << " FPS"<< endl;
        AccumTime = 0;
    }
    // Anima cubos
    AngY++;
    // Salva o tempo para o prximo ciclo de rendering
    last_idle_time = time_now;
    // Redesenha
    glutPostRedisplay();
}

// **********************************************************************
//  void keyboard ( unsigned char key, int x, int y )
//
//
// **********************************************************************
void keyboard ( unsigned char key, int x, int y )  
{
	switch ( key ) 
	{
        case 27:        // Termina o programa qdo
            exit ( 0 );   // a tecla ESC for pressionada
        break;
        case ' ':
            WireFrame = !WireFrame;
            break;
        default:
            cout << key;
      break;
  }
}

// **********************************************************************
//  void arrow_keys ( int a_keys, int x, int y )  
//
//
// **********************************************************************
void arrow_keys ( int a_keys, int x, int y )  
{
	switch ( a_keys ) 
	{
		case GLUT_KEY_UP:       // When Up Arrow Is Pressed...
			glutFullScreen ( ); // Go Into Full Screen Mode
			break;
	    case GLUT_KEY_DOWN:     // When Down Arrow Is Pressed...
			glutInitWindowSize  ( 700, 500 ); 
			break;
		default:
			break;
	}
}

// **********************************************************************
//  void main ( int argc, char** argv )
//
//
// **********************************************************************
int main ( int argc, char** argv )   
{
	glutInit            ( &argc, argv ); 
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
	glutInitWindowPosition (0,0);
	glutInitWindowSize  ( 700, 700 );
	glutCreateWindow    ( "Computacao Grafica - Exemplo de Geracao de Terrenos." );
    
    
    cout << "Pressione ESPACO!" << endl;
	init ();
    //system("pwd");
	
	glutDisplayFunc ( display );  
	glutReshapeFunc ( reshape );
	glutKeyboardFunc ( keyboard );
	glutSpecialFunc ( arrow_keys );
	glutIdleFunc ( animate );

	glutMainLoop ( );          
	return 0; 
}



